class DeferredLocator(object):
    """
    Create a deferred locator that can be used in matching
    elements in wait conditions.
    """
    def __init__(self, germanium):
        self._germanium = germanium

    def __call__(self):
        return self.element()

    def element(self, only_visible=True):
        """ Return the current matched element.
        :param only_visible: bool Find the element only
        if it's visible, so it can be interacted with.
        """
        if only_visible:
            elements = self.element_list(only_visible=only_visible)
            if not elements:
                return None

            return elements[0]

        return self._find_element()

    def element_list(self, only_visible=True):
        """
        :type only_visible: boolean
        """
        if not only_visible:
            return self._find_element_list()

        found_items = self._find_element_list()

        if not found_items:
            return list()

        def is_displayed_filter(item):
            try:
                return item.is_displayed()
            except Exception:
                return False

        return list(filter(is_displayed_filter, found_items))

    def exists(self, only_visible=True):
        """ Return True/False if the currently matched element exists or not
        :param only_visible:
        """
        return self.element(only_visible=only_visible) is not None

    def not_exists(self, only_visible=True):
        """
        :param only_visible:
        :return: True if the element is not existing/
        """
        return self.element(only_visible=only_visible) is None

    def _find_element(self):
        """ Find the element. """
        raise Exception("not implemented")

    def _find_element_list(self):
        """
        Find all the elements that match the given locator.
        :return:
        """
        raise Exception("Not implemented")



from .Element import Element


class TableCell(Element):
    def __init__(self, contains_text=None, *argv, **kw):
        super(TableCell, self).__init__("td", contains_text=contains_text, *argv, **kw)




from time import sleep
from germanium.impl import _ensure_list


def wait(closures, *extra_closures, while_not=None, timeout=10):
    """
    Executes a function given as argument every 400 milliseconds until it returns true. If it goes more than
    the timeout seconds, then this function throws an exception. If the function throws an exception, then
    it is assumed it is false.
    :param closures:
    :param while_not:
    :param timeout:
    :param extra_closures:
    """
    while_not = _ensure_list(while_not)
    closures = list(_ensure_list(closures))
    closures.extend(extra_closures)

    def closure_try_catch():
        for closure in closures:
            try:
                result = closure()
                if result:
                    return result
            except Exception as e:
                print("WARNING: waiting as false since: %s" % e)

    passed_timeout = 0
    closure_result = False

    while passed_timeout < timeout:
        for while_not_closure in while_not:
            try:
                if while_not_closure():
                    raise Exception("Waiting failed, since while_not condition matched")
            except Exception as e:
                raise Exception("Waiting failed, since while_not condition raised exception", e)

        closure_result = closure_try_catch()

        if closure_result:
            break

        passed_timeout += 0.4

        if passed_timeout >= 2:
            print("Running takes more than 2 seconds.")

        sleep(0.4)

    if not closure_result:
        raise Exception("Waiting has failed")




def _ensure_list(item):
    """
    Ensures that the given item is a list.
    :param item:
    :return:
    """
    if item is None:
        return []

    if isinstance(item, list):
        return item

    if isinstance(item, tuple):
        return list(item)

    return [item]



from .AbstractSelector import AbstractSelector


class Text(AbstractSelector):
    """
    Just a selector that finds the text in the page.
    """
    def __init__(self, text):
        super(AbstractSelector, self).__init__()
        self._text = text

    def get_selectors(self):
        """ Return the simple selector to find the text """
        js_selector = """js:
            var searchedText = "%s";

            function text(node) {
                return node.innerText || node.textContent || "";
            }

            if (text(document.body).indexOf(searchedText) < 0) {
                return null;
            }

            var processing_queue = [ document.body ];
            var result = [];

            while (processing_queue.length) {
                var foundChildElement = false;
                var currentNode = processing_queue.splice(0, 1)[0];

                for (var i = 0; i < currentNode.children.length; i++) {
                    if (text(currentNode.children[i]).indexOf(searchedText) >= 0) {
                        foundChildElement = true;
                        processing_queue.push(currentNode.children[i]);
                    }
                }

                if (!foundChildElement) {
                    result.push(currentNode);
                }
            }

            return result;
        """ % str(self._text).replace('"', '\\"', 100000)

        return [js_selector]
